from dataclasses import Field, fields
from inspect import getmodule
from textwrap import dedent
from typing import Type

from pytest_embrace.anno import AnnotationMap, get_pep_593_values

from .embrace import registry
from .exc import EmbraceError


class EmbraceTestGenError(EmbraceError):
    ...


def _stringify_type(type: Type) -> str:
    return str(type) if not repr(type).startswith("<class") else type.__name__


def _field_to_cute_type_hint(field: Field, anno_map: AnnotationMap) -> str:
    type_hint = _stringify_type(field.type)
    if field.name not in anno_map:
        return f"{field.name}: {type_hint}"

    anno_info = anno_map[field.name]
    comment = next((v for v in anno_info.annotations if isinstance(v, str)), None)
    type_hint = _stringify_type(anno_info.type)
    text = f"{field.name}: {type_hint}"
    if comment is not None:
        text += f"  # {comment}"
    return text


def gen_text(name: str, table: bool = False) -> str:
    case_type = registry().get(name)
    if case_type is None:
        raise EmbraceTestGenError(f"No such test type '{name}'.")

    anno_map = get_pep_593_values(case_type)

    type_hints = "\n".join(
        _field_to_cute_type_hint(f, anno_map) for f in fields(case_type)
    )

    source = getmodule(case_type)
    case_import = (
        dedent(
            f"""
            from {mod_name} import {case_type.__name__}
            """
        )
        if (mod_name := getattr(source, "__name__", None)) is not None
        else ""
    )

    # not using dedent bc newlines in the type hints are hard
    return f"""
from pytest_embrace import CaseArtifact
{case_import}

{type_hints}


def test({name}: CaseArtifact[{case_type.__name__}]) -> None:
    ...
            """



__version__ = "2.3.0"


from .case import CaseArtifact as CaseArtifact
from .case import derive_from_filename as derive_from_filename
from .case import trickles as trickles
from .embrace import Embrace as Embrace

# defining __all__ for the sake of generating docs with pdoc
__all__ = ["CaseArtifact", "derive_from_filename", "trickles", "Embrace"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_embrace']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.1,<2.0.0', 'pyperclip>=1.8.2,<2.0.0', 'pytest>=7.0,<8.0']

extras_require = \
{':python_version == "3.8"': ['typing-extensions>=4.2.0,<5.0.0']}

entry_points = \
{'console_scripts': ['embrace = pytest_embrace.cli:main'],
 'pytest11': ['pytest_embrace = pytest_embrace.plugin']}

setup_kwargs = {
    'name': 'pytest-embrace',
    'version': '2.3.0',
    'description': '💝  Dataclasses-as-tests. Describe the runtime once and multiply coverage with no boilerplate.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/pytest-embrace.svg)](https://badge.fury.io/py/pytest-embrace) [![PyPI pyversions](https://img.shields.io/pypi/pyversions/pytest-embrace.svg)](https://pypi.python.org/pypi/pytest-embrace/) ![CI](https://github.com/ainsleymcgrath/pytest-embrace/actions/workflows/ci.yml/badge.svg)\n\n![logotype](https://raw.githubusercontent.com/ainsleymcgrath/pytest-embrace/master/docs/logotype.svg)\n\nThe `pytest-embrace` plugin enables judicious, repeatable, lucid unit testing using.\n\nCheck out [the docs](https://ainsleymcgrath.github.io/pytest-embrace/) for more.\n',
    'author': 'Ainsley McGrath',
    'author_email': 'mcgrath.ainsley@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from importlib import reload
from textwrap import dedent

import pytest
from pyperclip import copy

from .embrace import registry
from .exc import EmbraceError
from .gen import gen_text
from .loader import load


# TODO entry, inj here
def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    registry_ = registry()
    embracers = [name for name in metafunc.fixturenames if name in registry_]

    if len(embracers) > 1:
        raise EmbraceError(f"Can't request multiple Embrace fixtures. Got {embracers}")

    if len(embracers) == 0:
        return

    (sut,) = embracers
    cls = registry_[sut]
    # ModuleInfo(case_name, case_cls, metafunc.module)
    cases = load(cls, metafunc.module)
    metafunc.parametrize("case", cases, ids=[str(c) for c in cases])

    reload(metafunc.module)  # this guarantees the safety of module scope.
    # by reloading at the end of a run, any future references to the just-tested module
    # will encounter it in its 'fresh' state.
    # see ../tests/test_plugin/test_safe_mutable_module_scope.py


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--embrace",
        dest="caller_fixture_name",
        help=dedent(
            "Takes the name of an Embrace() caller fixture and generates scaffolding"
            " for a new test module that will use it."
            "\nRun `pytest --embrace-ls` to see all the available caller fixtures."
        ),
    )
    parser.addoption(
        "--embrace-ls",
        help=dedent("List all the fixtures created for calling Embrace() test suites."),
        action="store_true",
    )


STOP_LOOP = object()  # for pytest hooks that stop on the first-non-None-result


def pytest_runtestloop(session: pytest.Session) -> object:
    generate_for = session.config.getoption("--embrace")
    do_ls = session.config.getoption("--embrace-ls")

    if generate_for is None and do_ls is None:
        return None  # allow normal pytest to happen

    reg = registry()

    if generate_for is not None:
        if generate_for not in reg:
            pytest.exit(
                f"No such fixture '{generate_for}'."
                f" Your options are {sorted([*reg])}"
            )

        # XXX inj here
        copypasta = gen_text(generate_for)
        print(f"\nCopying the following output to your clipboard:\n{copypasta}")
        copy(copypasta)
        return STOP_LOOP

    if do_ls is True:
        print()
        print(f"The following {len(reg)} Embrace() fixtures are abailable:\n")
        print("\n".join(f"{name}  (via {cls})" for name, cls in reg.items()))

        return STOP_LOOP

    return None
